from __future__ import absolute_import
input_name = '../examples/linear_elasticity/elastodynamic.py'
output_name_trunk = 'test_elastodynamic'

from tests_basic import TestInputEvolutionary

class Test(TestInputEvolutionary):
    pass
